from __future__ import annotations
from typing import TYPE_CHECKING

from CellFrame.Chain import ChainAtomPtr, ChainAddr
from CellFrame.Consensus import DAG, Block

if TYPE_CHECKING:
    from .net import CFChain
from .datums import CFDatum
from .crypto import CFSign
from ..common.types import ChainTypes


class CFBlock:
    def __init__(self, atom: ChainAtomPtr, chain: CFChain):
        self._origin_atom = atom
        if chain.type != ChainTypes.esbocs:
            raise TypeError(f"Chain type={chain.type} does not contain blocks")
        block = Block.fromAtom(self._origin_atom)
        if not block:
            raise RuntimeError(f"Not found block from atom")
        self._origin_block = block

        self.chain = chain
        self.hash = str(block.hash)
        self.created_at = block.created
        self.version = block.version
        self.cell_id = str(block.cellId)
        # TODO: move chain_id CFBlock --> CFChain
        self.chainid = str(block.chainId)
        self.signs = [CFSign(sign, self.chain.net) for sign in block.signs]
        self.meta = block.metaData

    def get_datums(self) -> list[CFDatum]:
        if self._origin_block.datums is None:
            return []
        return [CFDatum(self, datum) for datum in self._origin_block.datums]


class CFEvent:
    def __init__(self, atom: ChainAtomPtr, chain: CFChain):
        self._origin_atom = atom
        if chain.type != ChainTypes.dag_poa:
            raise TypeError(f"Chain type={chain.type} does not contain events")
        event = DAG.fromAtom(self._origin_atom)
        if not event:
            raise RuntimeError(f"Not found event from atom")
        self._origin_event = event

        self.chain = chain
        self.hash = str(event.hash)
        self.created_at = event.created
        self.version = event.version
        self.cell_id = str(event.cellId)
        self.signs = [CFSign(sign, self.chain.net) for sign in event.signs]
        self.round_id = event.roundId
        self.hashes = list(map(str, event.links))

    def get_datums(self) -> list[CFDatum]:
        # Event has only one Datum
        return [CFDatum(self, self._origin_event.datum)]
