from typing import Callable, TypeVar, Any, Protocol
from abc import ABC, abstractmethod, abstractproperty
from CellFrame.Chain import GlobalDB
from .datums import CFSign
from .net import CFNet


class GDBGroupBase(ABC):
    def __init__(self, group):
        self.group = group

    @abstractmethod
    def set(self, key: str, value: Any) -> None:
        pass

    @abstractmethod
    def get(self, key: str, net: CFNet, default=None) -> Any:
        pass

    @abstractmethod
    def delete(self, key: str) -> None:
        pass

    @abstractmethod
    def group_list(self) -> list[dict[str, Any]]:
        pass

    @staticmethod
    def add_to_white_list(net_name: str, group_mask: str, callback: Callable, callback_args: list):
        """
        group_mask example: "raiden.emission-center.*
        """
        GlobalDB.addSyncExtraGroup(net_name, group_mask, callback, callback_args)


D = TypeVar('D')


class GDBGroup(GDBGroupBase):
    def set(self, key: str, value: bytes) -> bool:
        return GlobalDB.set(key, self.group, value)

    def get(self, key: str, net: CFNet = None, default: D = None) -> bytes | D:
        return GlobalDB.get(key, self.group) or default

    def delete(self, key: str):
        GlobalDB.delete(key, self.group)

    def group_list(self) -> list[dict[str, bytes]]:
        return [{element.key: element.value} for element in GlobalDB.grLoad(self.group)]


class SerializableObject(Protocol):
    def serialize(self) -> bytes:
        pass

    def deserialize(self) -> object:
        pass


class GDBGroupSign(GDBGroupBase):
    def set(self, key: str, value: SerializableObject) -> bool:
        return GlobalDB.set(key, self.group, value.serialize())

    def get(self, key: str, net: CFNet, default: D = None) -> SerializableObject | D:
        data = GlobalDB.get(key, self.group)
        return CFSign.deserialize(data, net) or default

    def delete(self, key: str):
        GlobalDB.delete(key, self.group)

    def group_list(self) -> list[dict[str, bytes]]:
        return [{element.key: element.value} for element in GlobalDB.grLoad(self.group)]

# class GDBGroupObject(GDBGroupBase):
#     def set(self, key: str, value: SerializableObject) -> bool:
#         return GlobalDB.set(key, self.group, value.serialize())
#
#     def get(self, key: str, default: D = None) -> SerializableObject | D:
#         return GlobalDB.get(key, self.group).deserialize() or default
#
#     def delete(self, key: str):
#         GlobalDB.delete(key, self.group)
#
#     def group_list(self) -> list[dict[str, bytes]]:
#         return [{element.key: element.value} for element in GlobalDB.grLoad(self.group)]

# class GDBAtom(GDBGroup):
#     def set(self, key, value: CFAtom):
#         ...
