from typing import Literal, Callable

from .net import CFNet, MempoolCallback


class CFNets:
    def __init__(self, nets: list[CFNet]):
        self.nets = nets

    def mempool_notificator(self, *args, chain_name: Literal["main", "zerochain", "all"] = "all", **kwargs):
        """
        Usage example:
        @CFNets(NETS).mempool_notificator("to_args", chain_name="main", key="to_kwargs")
        def on_mempool_change(op_code: Literal["a", "d"], datum: CFDatum | datum_hash, *args, chain: 'CFChain', **kwargs):
            pass
        """

        def wrapper(callback: MempoolCallback):
            for net in self.nets:
                if chain_name == "all":
                    for chain in net.chains:
                        chain.register_mempool_notification_callback(callback, *args, **kwargs)
                else:
                    getattr(net, chain_name).register_mempool_notification_callback(callback, *args, **kwargs)

        return wrapper

    def atom_notificator(self, *args, chain_name: Literal["main", "zerochain", "all"] = "all", **kwargs):
        """
        Usage example:
        @CFNets(NETS).atom_notificator("to_args", chain_name="main", key="to_kwargs")
        def on_new_atom(atom: CFBlock | CFEvent, size: int, *args, chain: CFChain, **kwargs):
            pass
        """

        def wrapper(callback: Callable[[], None]):
            for net in self.nets:
                if chain_name == "all":
                    for chain in net.chains:
                        chain.register_atom_notification_callback(callback, *args, **kwargs)
                else:
                    getattr(net, chain_name).register_atom_notification_callback(callback, *args, **kwargs)

        return wrapper

    def gdbsync_notificator(self, *args, **kwargs):
        """
        Usage example:
        @CFNets(NETS).gdbsync_notificator("to_args", key="to_kwargs")
        def on_new_table_record(op_code, group, key, value, *args, net:CFNet, **kwargs):
            pass
        """

        def wrapper(callback: Callable[[], None]):
            for net in self.nets:
                net.register_gdbsync_notification_callback(callback, *args, **kwargs)

        return wrapper

    def ledger_tx_notificator(self, *args, **kwargs):
        """
        Usage example:
        @CFNets(NETS).ledger_tx_notificator("to_args", key="to_kwargs")
        def on_new_ledger_transaction(ledger, tx, *args, net: CFNet, **kwargs):
            pass
        """

        def wrapper(callback: Callable[[], None]):
            for net in self.nets:
                ledger = net.get_ledger()
                ledger.register_ledger_tx_notification_callback(callback, *args, **kwargs)

        return wrapper
